from collections import OrderedDict
from imutils import face_utils
import dlib
import cv2
import numpy as np
from .print_utils import show, write_on_image 

def draw_calc_brows(clone, shape, name, lst, iterList):
  step = 0.05 * clone.shape[0]
  i, j = lst
  write_on_image(clone, name,  (10, 30 + int(iterList["iter_brows"] * step)))

  # print('WHAT IS I, J: NAME', i, j, name)
  first_x, first_y = shape[i]
  last_x, last_y = shape[j-1]
  first_dim = np.array((first_x, first_y))
  last_dim = np.array((last_x, last_y))
  distance = np.linalg.norm(first_dim - last_dim)
  # print(first_dim, "\n", last_dim, '\n distance: ', distance)
  cv2.line(clone, first_dim, last_dim, (0,0,255), 1)
  write_on_image(clone, f'{distance:.2f}',  (clone.shape[1] // 3, 30 + int(iterList["iter_brows"] * step)))
  # show(clone)
  iterList["iter_brows"] += 1
  return distance



import numpy as np
import cv2

def gamma_correction(image, gamma):
    # print(f'gamma correction : {gamma}')
    # build a lookup table mapping the pixel values [0, 255] to
    # their adjusted gamma values
    invGamma = 1.0 / gamma
    table = np.array([((i / 255.0) ** invGamma) * 255
        for i in np.arange(0, 256)]).astype("uint8")
    # apply gamma correction using the lookup table
    return cv2.LUT(image, table)


import argparse
from .detectFacialFeatures import process_mshape
import cv2
import numpy as np

from pathlib import Path

SRC_DIR = Path(__file__).resolve().parent.parent

def main(args):
  process_mshape(args.input, args.m_shape, args.config_dat_path)

def predict_m_shape(front_face_img):
  from loguru import logger 
  logger.error("PREDICTING M SHAPE")
  # TODO
  front_face_img = front_face_img.convert('RGB') 

  front_face_img = np.array(front_face_img) 
  # Convert RGB to BGR 
  front_face_img = front_face_img[:, :, ::-1].copy() 

  cv2.imwrite("front_face_img.jpg", front_face_img)
  return process_mshape("front_face_img.jpg", True, str(SRC_DIR / 'm_shape/utilities/shape_predictor_81_face_landmarks.dat'))

if __name__ == "__main__":
  parser = argparse.ArgumentParser(description='MSHAPE and RECEDING HAIRLINE project')
  parser.add_argument('--single', type=str, default=False)
  # parser.add_argument('--single', type=str, default=r'C:\Users\Public\ferdy\front_face\black\male_black_hair\male_black_hair10.jpg', type=str)
  # TODO
  parser.add_argument('--input', default=str(SRC_DIR / 'sample_images/male_black_hair0.jpg'), type=str)
  parser.add_argument('--m_shape', default=True, type=str)
  parser.add_argument('--verbose', default=False, type=str)
  # TODO
  parser.add_argument('--config_dat_path', default=str(SRC_DIR / 'm_shape/utilities/shape_predictor_81_face_landmarks.dat'), type=str)
  args=parser.parse_args()
  main(args)



from collections import OrderedDict
from imutils import face_utils
import dlib
import cv2
import numpy as np 

print(__name__)
from ..utilities.drawAndCalculate import draw_calc_brows
from ..utilities.print_utils import show, return_stats
from ..utilities.algorithm_utils import perform_top_operations, perform_bottom_operations

person = 1
def process_mshape(image_path, m_shape, predictor):
  global person
  detector = dlib.get_frontal_face_detector()
  predictor = dlib.shape_predictor(predictor)

  image = cv2.imread(image_path)
  gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
      
  # Get faces into webcam's image
  rects = detector(gray, 0)
  
  coordinates = {
    "left_brow" : 0,
    "right_brow" : 0,
    "top_central_forehead_coor" : 0,
    "bottom_central_forehead_coor" : 0,
    "top_right_forehead_coor" : 0,
    "bottom_right_forehead_coor" : 0,
    "top_left_forehead_coor" : 0,
    "bottom_left_forehead_coor" : 0,
    "nose_central_coor" : 0,
    "chin_central_coor" : 0,
  }
  FACIAL_LANDMARKS_IDXS = OrderedDict([
      ("left_eyebrow", (17, 22)),
      ("right_eyebrow", (22, 27)),
      ("top_right_forehead_coor", (72)), 
      ("bottom_right_forehead_coor", (24)), 
      ("top_left_forehead_coor", (69)), 
      ("bottom_left_forehead_coor", (18)), 
      ("top_central_forehead_coor", (71)), 
      ("bottom_central_forehead_coor", (21)), 
      ("nose_central_coor", (33)), 
      ("chin_central_coor", (8)), 
    ])
  
  iterList = {
    "iter_brows": 0, 
    "iter_forehead": 0
  }

  # For each detected face, find the landmark.
  for (i, rect) in enumerate(rects):
      print("\n",6* "=", f"PERSON: {person} PATH: {image_path}",6 * "=")
      # Make the prediction and transfom it to numpy array
      shape = predictor(gray, rect)
      shape = face_utils.shape_to_np(shape)

      '''
      OUTPUT FOR EVERY FACE IS IN CLONE
      '''      
      clone = image.copy() 

      distance = {
        "forehead_central_distance":0,
        "forehead_left_distance":0,
        "forehead_right_distance":0,
        "left_brow":0,
        "right_brow":0,
        "forehead2nose": 0, 
        "nose2chin": 0
      }
      
      '''
      NOTES FOR PERFORM_operation: 
      we pass coordinates are part of mutable object, so when our algorithm runs, the changes of value in coordinates can be reflected. 
      since we already initialize the keys for the dictionary, updating or reassigning the value will return as pass by reference.
      '''

      for (name, lst) in FACIAL_LANDMARKS_IDXS.items():
        if name == 'left_eyebrow': 
          distance["left_brow"] = draw_calc_brows(clone, shape, name, lst, iterList)
        elif name == 'right_eyebrow': 
          distance["right_brow"] = draw_calc_brows(clone, shape, name, lst, iterList)
        elif name == 'nose_central_coor': 
          coordinates["nose_central_coor"] = np.array((shape[lst][0], shape[lst][1]))
        elif name == 'chin_central_coor': 
          coordinates["chin_central_coor"] = np.array((shape[lst][0], shape[lst][1]))
        elif name.startswith("top"):
          perform_top_operations(shape, lst, image, name, coordinates)
        elif name.startswith("bottom"):
          perform_bottom_operations(shape, lst, clone, name, coordinates, distance, iterList)
  '''
  detect m shape
  '''
  if m_shape == True:
    print(f'WHAT IS MSHAPE: {m_shape}')
    results = return_stats(distance["forehead_central_distance"],distance["forehead_left_distance"], distance["forehead_right_distance"], m_shape=True)
    print(results)
    return results
    # with open(f'./results/{person}.txt', 'w') as f: 
    #   f.write(results)
  # else:
  #   distance["forehead2nose"] =  np.linalg.norm(coordinates["bottom_central_forehead_coor"] - coordinates["nose_central_coor"])
  #   # cv2.line(clone, coordinates["bottom_central_forehead_coor"], coordinates["nose_central_coor"], (0,255,0), 2)
  #   distance["nose2chin"] = np.linalg.norm(coordinates["nose_central_coor"] - coordinates["chin_central_coor"])
    # cv2.line(clone, coordinates["nose_central_coor"], coordinates["chin_central_coor"], (255,0,), 2)

    # results = return_stats(distance["forehead_central_distance"],distance["forehead2nose"], distance["nose2chin"], m_shape=False)
    # print(results)
    # with open(f'./results/{person}.txt', 'w') as f: 
    #   f.write(results)

  # show(clone)
  # cv2.imwrite(f'./results/{person}.jpg', clone)

  # person += 1

  # clean up iterator for next image
  for key in iterList.keys(): 
    iterList[key] = 0
 




        


import numpy as np  
import cv2
from .preprocess import gamma_correction
from .print_utils import print_training_stats, show, write_on_image

##################################################
############### REPEATED ALGORITHM  ###############
##################################################
def perform_bottom_operations(shape, lst, clone, name, coordinates, distance, iterList):
  step = 0.05 * clone.shape[0]
  _, first_y = shape[lst]
  position = name.split("_")[1]
  coordinates[f"{name}"]  = np.array((coordinates[f"top_{position}_forehead_coor"][0], first_y))
  cv2.rectangle(clone, coordinates[f"top_{position}_forehead_coor"], coordinates[f"{name}"], (0,0,255), -1)
  distance[f"forehead_{position}_distance"] = np.linalg.norm(coordinates[f"{name}"] - coordinates[f"top_{position}_forehead_coor"])
  write_on_image(clone, f'forehead_{position}:{distance[f"forehead_{position}_distance"]:.1f}', (clone.shape[1] // 2, 30 + int(iterList["iter_forehead"] * step)))
  iterList["iter_forehead"] += 1 

def perform_top_operations(shape, lst, image, name, coordinates):
  first_x, first_y = shape[lst]
  
  average = get_average(first_x, first_y,  detect_and_correct(image))

  # print(f'WHAT IS AVERAGE: {average}')

  first_x, first_y = training(first_x, first_y, average, image)

  coordinates[f"{name}"] = np.array((first_x, first_y))
  # print(f"what is the coordinate for {name}: ", coordinates[f"{name}"])

def training(first_x, first_y, average, image):
  iteratio = 1
  step = 0.01 * image.shape[0]
  while  average > 168 and iteratio < 6:
    first_y =int(first_y - step)
    iteratio +=1 
    cv2.circle(image, (first_x, first_y), 1, (0, 255, 0), -1)

    ROI = image[first_x, first_y]
    
    # print_training_stats(iteratio, first_y, ROI)
    '''
    during training no need to perform another gammna correction
    '''
    get_average(first_x, first_y, image)

    # show(image)
  
  return first_x, first_y
  

'''
@output: return image after performing gamma correciton
'''
def detect_and_correct(image):
  average = np.average(image)
  # print("average: ", "=" * 10, average)
  if average < 120: 
    # print("*" * 10, " increasing light")
    image = gamma_correction(image, 2.5)
  elif average > 200: 
    # print("*" * 10, " decreasing light")
    image =  gamma_correction(image, 0.5)
  return image


def get_average(first_x, first_y, image):        
  ROI = image[first_x-20: first_x+20, first_y, :]
  kernel = np.ones((40, 1,3))
  result = ROI * kernel
  average = np.average(result)
  # print("average: ", "=" * 10, average)
  return average



import cv2

##################################################
############### PRINTING UTILITIES ###############
##################################################

# def format_string(text): 
#   print("=" * 10, f' {text} ', "=" * 10, "\n")

def format_string(text): 
  return "=" * 10 +  f' {text} ' +  "=" * 10 +  "\n"

# def print_stats(top_bottom_distance, left, right, m_shape=False):
#   measurement_type = 'mshape' if m_shape else 'forehead distance'

#   format_string(f'PRINT STATS FOR {measurement_type}')

#   print(f'forehead central distance: {top_bottom_distance}')

#   if m_shape:
#     print(f'left forehead distance: {left}')
#     print(f'right forehead distance: {right}')
#     total_dist = left + right + top_bottom_distance
#     left = left/ total_dist
#     top_bottom_distance = top_bottom_distance/ total_dist
#     right = right/ total_dist
#     print(f'RATIO: \n LEFT: {left} CENTRAL: {top_bottom_distance:.2f} : RIGHT: {right}')


#   if not m_shape:
#     print(f'left brow distance: {left}')
#     print(f'right brow distance: {right}')
#     print(f'ratio forehead line to left eyebrow: {(top_bottom_distance / left):.2f}')
#     print(f'ratio forehead line to right eyebrow: {(top_bottom_distance / right):.2f}')
#   print("\n")

def return_stats(top_bottom_distance, left, right, m_shape=False):
  text = ""
  measurement_type = 'M-SHAPE' if m_shape else 'FNC (FOREHEAD-NOSE-CHIN)'

  text += format_string(f'{measurement_type} PROJECT')

  text += (f'forehead central distance: {top_bottom_distance} \n')

  if m_shape:
    text += (f'left forehead distance: {left}\n')
    text += (f'right forehead distance: {right}\n')
    total_dist = left + right + top_bottom_distance
    left = left/ total_dist
    top_bottom_distance = top_bottom_distance/ total_dist
    right = right/ total_dist

    text += (f'{format_string("RATIO")}LEFT: {left}\nCENTRAL: {top_bottom_distance:.2f}\nRIGHT: {right}')

    print(left)
    diag = 0
    if left> 0.39:
      diag += 1
    if right> 0.39:
      diag += 1
    if top_bottom_distance < left and top_bottom_distance < right:
      diag += 1
    if diag == 3: 
      diag = 'yes'
    else: 
      diag = 'no' 

    return [left, top_bottom_distance, right, diag]


  if not m_shape:
    # text += (f'left brow distance: {left}\n')
    # text += (f'right brow distance: {right}\n')
    # text += (f'ratio forehead line to left eyebrow: {(top_bottom_distance / left):.2f}')
    # text += (f'ratio forehead line to right eyebrow: {(top_bottom_distance / right):.2f}')
    text += (f'forehead2nose distance: {left}\n')
    text += (f'nose2chin: {right}\n')
    total_dist = left + right + top_bottom_distance
    left = left/ total_dist
    top_bottom_distance = top_bottom_distance/ total_dist
    right = right/ total_dist
    text += (f'RATIO: \nFOREHEAD2NOSE: {left} FOREHEAD: {top_bottom_distance:.2f} : NOSE2CHIN: {right}')
  return text

def print_training_stats(iteratio, first_y, ROI):
  print(3 * "*", f'iter: {iteratio}', 3 * "*")
  print("WHAT IS FIRST Y: ", "=" * 20, first_y)

  print('extending forehead line')

  print('WHAT IS ROI: ',ROI)

##################################################
############### OPEN CV DRAW AND SHOW UTILITIES ###############
##################################################

def write_on_image(clone, text, position): 
  cv2.putText(clone, f'{text}', position, cv2.FONT_HERSHEY_SIMPLEX, 0.5 * clone.shape[1] / clone.shape[0], (0, 0, 255), 2)

def show(image):
  cv2.imshow("image",image)
  cv2.waitKey(0)
  cv2.destroyAllWindows()

